import pytest
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import math
import matplotlib.pyplot as plt
from train_methods import SORNet, SOR_CHEBY_Net, AORNet, RINet

# Model parameters and other settings (remain the same as in the previous code)

# ... (The rest of the code remains the same)

# Test functions
def test_SORNet_parameters():
    # Test parameters of SORNet
    num_itr = 25
    bs = 10
    y = torch.ones(bs, m).to(device)

    model_SorNet = SORNet(num_itr).to(device)
    assert model_SorNet.inv_omega.shape == torch.Size([1])
    assert model_SorNet.inv_omega[0] == init_val_SORNet

def test_SOR_CHEBY_Net_parameters():
    # Test parameters of SOR_CHEBY_Net
    num_itr = 25
    bs = 10
    y = torch.ones(bs, m).to(device)

    model_Sor_Cheby_Net = SOR_CHEBY_Net(num_itr).to(device)
    assert model_Sor_Cheby_Net.gamma.shape == torch.Size([num_itr])
    assert model_Sor_Cheby_Net.gamma[0] == init_val_SOR_CHEBY_Net_omega
    assert model_Sor_Cheby_Net.omega.shape == torch.Size([num_itr])
    assert model_Sor_Cheby_Net.omega[0] == init_val_SOR_CHEBY_Net_gamma
    assert model_Sor_Cheby_Net.inv_omega.shape == torch.Size([1])
    assert model_Sor_Cheby_Net.inv_omega[0] == init_val_SOR_CHEBY_Net_alpha

def test_AORNet_parameters():
    # Test parameters of AORNet
    num_itr = 25
    bs = 10
    y = torch.ones(bs, m).to(device)

    model_AorNet = AORNet(num_itr).to(device)
    assert model_AorNet.r.shape == torch.Size([1])
    assert model_AorNet.r[0] == init_val_AORNet_r
    assert model_AorNet.omega.shape == torch.Size([1])
    assert model_AorNet.omega[0] == init_val_AORNet_omega

def test_RINet_parameters():
    # Test parameters of RINet
    num_itr = 25
    bs = 10
    y = torch.ones(bs, m).to(device)

    rinet_model = RINet(num_itr).to(device)
    assert rinet_model.inv_omega.shape == torch.Size([1])
    assert rinet_model.inv_omega[0] == init_val_RINet

# To run all the tests at once, you can call the following function
# This will execute all the individual test functions
def test_all_models_parameters():
    test_SORNet_parameters()
    test_SOR_CHEBY_Net_parameters()
    test_AORNet_parameters()
    test_RINet_parameters()

# Call the test function
test_all_models_parameters()


import pytest
import numpy as np
import math
from Iterative import gs_model, ri_model, Jacobi_model, sor_model, sor_cheby_model, aor_model, aor_cheby_model, solution
import torch

device = torch.device('cpu') # 'cpu' or 'cuda'
## model parameters 
itr = 25 # iteration steps $T$
num_itr = itr # max. iterations
n = 300 
m = 600 
##


## parameters for the evaluation of generalization error 
total_itr = 25 # total number of iterations (multiple number of "itr")
bs = 10 # number of samples

# Individual test functions

def test_gs_model():
    # define test parameters
    num_itr = 25 # iteration steps $T$
    bs = 10 # number of samples
    y = torch.ones(bs, m).to(device)

    # test GS
    s, traj = gs_model(num_itr, bs, y)
    assert s.shape == (bs, n)
    assert len(traj) == num_itr + 1

def test_ri_model():
    # define test parameters
    num_itr = 25 # iteration steps $T$
    bs = 10 # number of samples
    y = torch.ones(bs, m).to(device)

    # test RI
    s, traj = ri_model(num_itr, bs, y)
    assert s.shape == (bs, n)
    assert len(traj) == num_itr + 1

def test_Jacobi_model():
    # define test parameters
    num_itr = 25 # iteration steps $T$
    bs = 10 # number of samples
    y = torch.ones(bs, m).to(device)

    # test Jacobi
    s, traj = Jacobi_model(num_itr, bs, y)
    assert s.shape == (bs, n)
    assert len(traj) == num_itr + 1

def test_sor_model():
    # define test parameters
    num_itr = 25 # iteration steps $T$
    bs = 10 # number of samples
    y = torch.ones(bs, m).to(device)

    # test SOR
    s, traj = sor_model(num_itr, bs, y)
    assert s.shape == (bs, n)
    assert len(traj) == num_itr + 1


# Copyright (c) 2022-2023 Salah Berra and contributors
# Distributed under the the GNU General Public License (See accompanying file LICENSE or copy
# at https://www.gnu.org/licenses/)

import torch
import numpy as np
from typing import List, Tuple
from .utils import device, decompose_matrix

def model_iterations(total_itr: int, n: int, bs: int, model, solution: torch.Tensor) -> Tuple[List[torch.Tensor], List[float]]:
    """
    Perform iterations using the provided model and calculate the error norm at each iteration.

    Args:
        - total_itr (int): Total number of iterations to perform.
        - n (int): Dimension of the solution.
        - bs (int): Batch size.
        - model: The model instance that provides the iterate method.
        - solution (torch.Tensor): The ground truth solution tensor.

    Returns:
        - Tuple[List[torch.Tensor], List[float]]:
            - List of tensors representing the solution estimates at each iteration.
            - List of float values representing the normalized error at each iteration.
    """
    norm_list_model = []  # Initialize the iteration list
    s_hats = []
    
    for i in range(total_itr + 1):
        s_hat, _ = model.iterate(i)
        err = (torch.norm(solution.to(device) - s_hat.to(device)) ** 2).item() / (n * bs)
        
        s_hats.append(s_hat)
        norm_list_model.append(err)
    
    return s_hats, norm_list_model

class base_model:
    """
    Base model class for matrix decomposition and initialization.

    Args:
        - n (int): Dimension of the solution.
        - A (np.ndarray): Input square matrix to decompose.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.

    Attributes:
        - n (int): Dimension of the solution.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - A (torch.Tensor): Original matrix converted to a torch tensor.
        - D (torch.Tensor): Diagonal matrix of A.
        - L (torch.Tensor): Lower triangular matrix of A.
        - U (torch.Tensor): Upper triangular matrix of A.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - Minv (torch.Tensor): Inverse of the matrix (D + L).
    """

    def __init__(self, n: int, A: np.ndarray, H: torch.Tensor, bs: int, y: torch.Tensor):
        """
        Initialize the base_model with the given parameters and decompose matrix A.

        Args:
            - n (int): Dimension of the solution.
            - A (np.ndarray): Input square matrix to decompose.
            - H (torch.Tensor): Random matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution tensor.
        """
        self.n = n
        self.H = H
        self.bs = bs
        self.y = y

        self.A, self.D, self.L, self.U, self.Dinv, self.Minv = decompose_matrix(A)

class GS(base_model):
    """
    Class implementing the Gauss-Seidel algorithm for solving a linear system.

    Args:
        - n (int): Dimension of the solution.
        - A (np.ndarray): Input square matrix to decompose.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.

    Attributes:
        - n (int): Dimension of the solution.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - A (torch.Tensor): Original matrix converted to a torch tensor.
        - D (torch.Tensor): Diagonal matrix of A.
        - L (torch.Tensor): Lower triangular matrix of A.
        - U (torch.Tensor): Upper triangular matrix of A.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - Minv (torch.Tensor): Inverse of the matrix (D + L).
        - num_itr (int): The number of Gauss-Seidel iterations to perform.

    Methods:
        - iterate(num_itr: int) -> Tuple[torch.Tensor, list]:
            Performs the Gauss-Seidel iterations and returns the final solution and trajectory of solutions.
    """

    def __init__(self, n: int, A: np.ndarray, H: torch.Tensor, bs: int, y: torch.Tensor):
        """
        Initialize the Gauss-Seidel solver.

        Args:
            n (int): Dimension of the solution.
            A (np.ndarray): Input square matrix to decompose.
            H (torch.Tensor): Random matrix H.
            bs (int): Batch size.
            y (torch.Tensor): Solution tensor.
        """
        super(GS, self).__init__(n, A, H, bs, y)
        
    def iterate(self, num_itr: int) -> Tuple[torch.Tensor, list]:
        """
        Perform the Gauss-Seidel iterations and return the final solution.

        Args:
            - num_itr (int): The number of iterations to perform.

        Returns:
            - Tuple[torch.Tensor, list]: 
                - torch.Tensor: The final solution tensor of shape (bs, n).
                - list: List containing the trajectory of solutions throughout the iterations.
        """
        traj = []
        s = torch.zeros(self.bs, self.n).to(device)
        traj.append(s)

        yMF = torch.matmul(self.y, self.H.T)  # Assuming H is defined
        s = torch.matmul(yMF, self.Dinv)  # Generate batch initial solution vector

        for i in range(num_itr):
            temp = -torch.matmul(s, self.U) + yMF
            s = torch.matmul(temp, self.Minv)
            traj.append(s)

        return s, traj

class RI(base_model):
    """
    Class implementing the Richardson iteration algorithm for solving a linear system.

    Args:
        - n (int): Dimension of the solution.
        - A (np.ndarray): Input square matrix to decompose.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.

    Attributes:
        - n (int): Dimension of the solution.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - A (torch.Tensor): Original matrix converted to a torch tensor.
        - D (torch.Tensor): Diagonal matrix of A.
        - L (torch.Tensor): Lower triangular matrix of A.
        - U (torch.Tensor): Upper triangular matrix of A.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - Minv (torch.Tensor): Inverse of the matrix (D + L).

    Methods:
        iterate(num_itr: int) -> Tuple[torch.Tensor, list]:
            Performs the Richardson iterations and returns the final solution and trajectory of solutions.
    """

    def __init__(self, n: int, A: np.ndarray, H: torch.Tensor, bs: int, y: torch.Tensor):
        """
        Initialize the Richardson iteration solver.

        Args:
            - n (int): Dimension of the solution.
            - A (np.ndarray): Input square matrix to decompose.
            - H (torch.Tensor): Random matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution tensor.
        """
        super(RI, self).__init__(n, A, H, bs, y)

    def iterate(self, num_itr: int) -> Tuple[torch.Tensor, list]:
        """
        Perform the Richardson iterations and return the final solution.

        Args:
            num_itr (int): The number of iterations to perform.

        Returns:
            - Tuple[torch.Tensor, list]:
                - torch.Tensor: The final solution tensor of shape (bs, n).
                - list: List containing the trajectory of solutions throughout the iterations.
        """
        traj = []
        omega = torch.tensor(0.25)
        s = torch.zeros(self.bs, self.n).to(device)
        traj.append(s)

        yMF = torch.matmul(self.y, self.H.T)  # Assuming H is defined
        s = torch.matmul(yMF, self.Dinv)  # Generate batch initial solution vector

        for i in range(num_itr):
            s = s + torch.mul(omega, (yMF - torch.matmul(s, self.A)))
            traj.append(s)

        return s, traj

class Jacobi(base_model):
    """
    Class implementing the Jacobi iteration algorithm for solving a linear system.

    Args:
        - n (int): Dimension of the solution.
        - A (np.ndarray): Input square matrix to decompose.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - omega (float, optional): Relaxation parameter for Jacobi iterations. Defaults to 0.2.

    Attributes:
        - n (int): Dimension of the solution.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - A (torch.Tensor): Original matrix converted to a torch tensor.
        - D (torch.Tensor): Diagonal matrix of A.
        - L (torch.Tensor): Lower triangular matrix of A.
        - U (torch.Tensor): Upper triangular matrix of A.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - Minv (torch.Tensor): Inverse of the matrix (D + L).
        - omega (torch.Tensor): Relaxation parameter for Jacobi iterations.

    Methods:
        iterate(num_itr: int) -> Tuple[torch.Tensor, list]:
            Performs the Jacobi iterations and returns the final solution and trajectory of solutions.
    """

    def __init__(self, n: int, A: np.ndarray, H: torch.Tensor, bs: int, y: torch.Tensor, omega: float = 0.2):
        """
        Initialize the Jacobi iteration solver.

        Args:
            - n (int): Dimension of the solution.
            - A (np.ndarray): Input square matrix to decompose.
            - H (torch.Tensor): Random matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution tensor.
            - omega (float, optional): Relaxation parameter for Jacobi iterations. Defaults to 0.2.
        """
        super(Jacobi, self).__init__(n, A, H, bs, y)
        self.omega = torch.tensor(omega)

    def iterate(self, num_itr: int) -> Tuple[torch.Tensor, list]:
        """
        Perform the Jacobi iterations and return the final solution.

        Args:
            num_itr (int): The number of iterations to perform.

        Returns:
            - Tuple[torch.Tensor, list]:
                - torch.Tensor: The final solution tensor of shape (bs, n).
                - list: List containing the trajectory of solutions throughout the iterations.
        """
        traj = []
        s = torch.zeros(self.bs, self.n).to(device)
        traj.append(s)

        yMF = torch.matmul(self.y, self.H.T)  # Assuming H is defined
        s = torch.matmul(yMF, self.Dinv)  # Generate batch initial solution vector

        for i in range(num_itr):
            temp = torch.matmul(self.Dinv, (self.D - self.A))
            s = torch.matmul(s, temp) + torch.matmul(yMF, self.Dinv)
            traj.append(s)

        return s, traj

class SOR(base_model):
    """
    Class implementing the Successive Over-Relaxation (SOR) algorithm for solving a linear system.

    Args:
        - n (int): Dimension of the solution.
        - A (np.ndarray): Input square matrix to decompose.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - omega (float, optional): Relaxation parameter for SOR iterations. Defaults to 1.8.

    Attributes:
        - n (int): Dimension of the solution.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - A (torch.Tensor): Original matrix converted to a torch tensor.
        - D (torch.Tensor): Diagonal matrix of A.
        - L (torch.Tensor): Lower triangular matrix of A.
        - U (torch.Tensor): Upper triangular matrix of A.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - Minv (torch.Tensor): Inverse of the matrix (D + L).
        - omega (torch.Tensor): Relaxation parameter for SOR iterations.

    Methods:
        iterate(num_itr: int) -> Tuple[torch.Tensor, list]:
            Performs the SOR iterations and returns the final solution and trajectory of solutions.
    """

    def __init__(self, n: int, A: np.ndarray, H: torch.Tensor, bs: int, y: torch.Tensor, omega: float = 1.8):
        """
        Initialize the SOR solver.

        Args:
            - n (int): Dimension of the solution.
            - A (np.ndarray): Input square matrix to decompose.
            - H (torch.Tensor): Random matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution tensor.
            - omega (float, optional): Relaxation parameter for SOR iterations. Defaults to 1.8.
        """
        super(SOR, self).__init__(n, A, H, bs, y)
        self.omega = torch.tensor(omega)

    def iterate(self, num_itr: int) -> Tuple[torch.Tensor, list]:
        """
        Perform the SOR iterations and return the final solution.

        Args:
            num_itr (int): The number of iterations to perform.

        Returns:
            - Tuple[torch.Tensor, list]:
                - torch.Tensor: The final solution tensor of shape (bs, n).
                - list: List containing the trajectory of solutions throughout the iterations.
        """
        traj = []
        n = self.y.size(1)

        inv_omega = torch.div(1, self.omega)
        invM_sor = torch.linalg.inv(self.D - torch.mul(inv_omega, self.L))

        s = torch.zeros(self.bs, n).to(device)
        traj.append(s)

        yMF = torch.matmul(self.y, self.H.T)
        s = torch.matmul(yMF, self.Dinv)

        for i in range(num_itr):
            temp = torch.mul((inv_omega - 1), self.D) + torch.mul(inv_omega, self.U)
            s = torch.matmul(s, torch.matmul(invM_sor, temp)) + torch.matmul(yMF, invM_sor)
            traj.append(s)

        return s, traj


class SOR_CHEBY(base_model):
    """
    Class implementing the SOR-Chebyshev algorithm for solving a linear system.

    Args:
        - n (int): Dimension of the solution.
        - A (np.ndarray): Input square matrix to decompose.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - omega (float, optional): Relaxation parameter for SOR iterations. Defaults to 1.8.
        - omegaa (float, optional): Acceleration parameter for SOR-Chebyshev iterations. Defaults to 0.8.
        - gamma (float, optional): Damping factor for SOR-Chebyshev iterations. Defaults to 0.8.

    Attributes:
        - n (int): Dimension of the solution.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - A (torch.Tensor): Original matrix converted to a torch tensor.
        - D (torch.Tensor): Diagonal matrix of A.
        - L (torch.Tensor): Lower triangular matrix of A.
        - U (torch.Tensor): Upper triangular matrix of A.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - Minv (torch.Tensor): Inverse of the matrix (D + L).
        - omega (torch.Tensor): Relaxation parameter for SOR iterations.
        - omegaa (torch.Tensor): Acceleration parameter for SOR-Chebyshev iterations.
        - gamma (torch.Tensor): Damping factor for SOR-Chebyshev iterations.

    Methods:
        iterate(num_itr: int) -> Tuple[torch.Tensor, list]:
            Performs the SOR-Chebyshev iterations and returns the final solution and trajectory of solutions.
    """

    def __init__(self, n: int, A: np.ndarray, H: torch.Tensor, bs: int, y: torch.Tensor, omega: float = 1.8, omegaa: float = 0.8, gamma: float = 0.8):
        """
        Initialize the SOR-Chebyshev solver.

        Args:
            - n (int): Dimension of the solution.
            - A (np.ndarray): Input square matrix to decompose.
            - H (torch.Tensor): Random matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution tensor.
            - omega (float, optional): Relaxation parameter for SOR iterations. Defaults to 1.8.
            - omegaa (float, optional): Acceleration parameter for SOR-Chebyshev iterations. Defaults to 0.8.
            - gamma (float, optional): Damping factor for SOR-Chebyshev iterations. Defaults to 0.8.
        """
        super(SOR_CHEBY, self).__init__(n, A, H, bs, y)
        self.omega = torch.tensor(omega)
        self.omegaa = torch.tensor(omegaa)
        self.gamma = torch.tensor(gamma)

    def iterate(self, num_itr: int) -> Tuple[torch.Tensor, list]:
        """
        Perform the SOR-Chebyshev iterations and return the final solution.

        Args:
            num_itr (int): The number of iterations to perform.

        Returns:
            - Tuple[torch.Tensor, list]:
                - torch.Tensor: The final solution tensor of shape (bs, n).
                - list: List containing the trajectory of solutions throughout the iterations.
        """
        traj = []

        inv_omega = torch.div(1, self.omega)
        invM_sor = torch.linalg.inv(self.D - torch.mul(inv_omega, self.L))

        s = torch.zeros(self.bs, self.n).to(device)
        s_new = torch.zeros(self.bs, self.n).to(device)
        traj.append(s)

        yMF = torch.matmul(self.y, self.H.T)
        s = torch.matmul(yMF, self.Dinv)

        s_present = s
        s_old = torch.zeros(s_present.shape).to(device)

        for i in range(num_itr):
            temp = torch.mul((inv_omega - 1), self.D) + torch.mul(inv_omega, self.U)
            s = torch.matmul(s, torch.matmul(invM_sor, temp)) + torch.matmul(yMF, invM_sor)

            s_new = self.omegaa * (self.gamma * (s - s_present) + (s_present - s_old)) + s_old
            s_old = s
            s_present = s_new

            traj.append(s_new)

        return s_new, traj

class AOR(base_model):
    """
    Class implementing the Accelerated Over-Relaxation (AOR) algorithm for solving a linear system.

    Args:
        - n (int): Dimension of the solution.
        - A (np.ndarray): Input square matrix to decompose.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - omega (float, optional): Relaxation parameter for AOR iterations. Defaults to 0.3.
        - r (float, optional): Relaxation parameter r. Defaults to 0.2.

    Attributes:
        - n (int): Dimension of the solution.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - A (torch.Tensor): Original matrix converted to a torch tensor.
        - D (torch.Tensor): Diagonal matrix of A.
        - L (torch.Tensor): Lower triangular matrix of A.
        - U (torch.Tensor): Upper triangular matrix of A.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - Minv (torch.Tensor): Inverse of the matrix (D + L).
        - omega (torch.Tensor): Relaxation parameter for AOR iterations.
        - r (torch.Tensor): Relaxation parameter r.

    Methods:
        iterate(num_itr: int) -> Tuple[torch.Tensor, list]:
            Performs the AOR iterations and returns the final solution and trajectory of solutions.
    """

    def __init__(self, n: int, A: np.ndarray, H: torch.Tensor, bs: int, y: torch.Tensor, omega: float = 0.3, r: float = 0.2):
        """
        Initialize the AOR solver.

        Args:
            - n (int): Dimension of the solution.
            - A (np.ndarray): Input square matrix to decompose.
            - H (torch.Tensor): Random matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution tensor.
            - omega (float, optional): Relaxation parameter for AOR iterations. Defaults to 0.3.
            - r (float, optional): Relaxation parameter r. Defaults to 0.2.
        """
        super(AOR, self).__init__(n, A, H, bs, y)
        self.omega = torch.tensor(omega)
        self.r = torch.tensor(r)

    def iterate(self, num_itr: int) -> Tuple[torch.Tensor, list]:
        """
        Perform the AOR iterations and return the final solution.

        Args:
            num_itr (int): The number of iterations to perform.

        Returns:
            - Tuple[torch.Tensor, list]:
                - torch.Tensor: The final solution tensor of shape (bs, n).
                - list: List containing the trajectory of solutions throughout the iterations.
        """
        traj = []

        M = (self.D - torch.mul(self.r, self.L))
        invM_aor = torch.linalg.inv(M)
        N = (torch.mul((1 - self.omega), self.D) + torch.mul((self.omega - self.r), self.L) + torch.mul(self.omega, self.U))

        s = torch.zeros(self.bs, self.n).to(device)
        traj.append(s)

        yMF = torch.matmul(self.y, self.H.T)
        s = torch.matmul(yMF, self.Dinv)

        for i in range(num_itr):
            s = torch.matmul(s, torch.matmul(invM_aor, N)) + torch.mul(self.omega, torch.matmul(yMF, invM_aor))
            traj.append(s)

        return s, traj


class AOR_CHEBY(base_model):
    """
    Class implementing the Accelerated Over-Relaxation (AOR) with Chebyshev acceleration algorithm for solving a linear system.

    Args:
        - n (int): Dimension of the solution.
        - A (np.ndarray): Input square matrix to decompose.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - omega (float, optional): Relaxation parameter for AOR iterations. Defaults to 0.1.
        - r (float, optional): Relaxation parameter r. Defaults to 0.1.

    Attributes:
        - n (int): Dimension of the solution.
        - H (torch.Tensor): Random matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution tensor.
        - A (torch.Tensor): Original matrix converted to a torch tensor.
        - D (torch.Tensor): Diagonal matrix of A.
        - L (torch.Tensor): Lower triangular matrix of A.
        - U (torch.Tensor): Upper triangular matrix of A.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - Minv (torch.Tensor): Inverse of the matrix (D + L).
        - omega (torch.Tensor): Relaxation parameter for AOR iterations.
        - r (torch.Tensor): Relaxation parameter r.

    Methods:
        iterate(num_itr: int) -> Tuple[torch.Tensor, list]:
            Performs the AOR-Chebyshev iterations and returns the final solution and trajectory of solutions.
    """

    def __init__(self, n: int, A: np.ndarray, H: torch.Tensor, bs: int, y: torch.Tensor, omega: float = 0.1, r: float = 0.1):
        """
        Initialize the AOR-Chebyshev solver.

        Args:
            - n (int): Dimension of the solution.
            - A (np.ndarray): Input square matrix to decompose.
            - H (torch.Tensor): Random matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution tensor.
            - omega (float, optional): Relaxation parameter for AOR iterations. Defaults to 0.1.
            - r (float, optional): Relaxation parameter r. Defaults to 0.1.
        """
        super(AOR_CHEBY, self).__init__(n, A, H, bs, y)
        self.omega = torch.tensor(omega)
        self.r = torch.tensor(r)

    def iterate(self, num_itr: int) -> Tuple[torch.Tensor, list]:
        """
        Perform the AOR-Chebyshev iterations and return the final solution.

        Args:
            num_itr (int): The number of iterations to perform.

        Returns:
            - Tuple[torch.Tensor, list]:
                - torch.Tensor: The final solution tensor of shape (bs, n).
                - list: List containing the trajectory of solutions throughout the iterations.
        """
        traj = []

        s = torch.zeros(self.bs, self.n).to(device)
        traj.append(s)

        yMF = torch.matmul(self.y, self.H.T)
        s = torch.matmul(yMF, self.Dinv)

        Y0 = s

        M = (self.D - torch.mul(self.r, self.L))
        invM = torch.linalg.inv(M)
        N = (torch.mul((1 - self.omega), self.D) + torch.mul((self.omega - self.r), self.L) + torch.mul(self.omega, self.U))
        temp = torch.matmul(invM, N)

        rho = torch.tensor(0.1)
        mu0 = torch.tensor(1)
        mu1 = rho
        xhat1 = torch.matmul(s, temp) + self.omega * torch.matmul(yMF, invM)
        Y1 = xhat1
        Y = Y1

        for i in range(num_itr):
            f = 2 / (rho * mu1)
            j = 1 / mu0
            c = f - j
            mu = 1 / c
            a = (2 * mu) / (rho * mu1)
            Y = torch.matmul((Y1 * a), torch.matmul(invM, N)) - (((mu / mu0)) * Y0) + (a * torch.matmul(yMF, invM))
            Y0 = Y1
            Y1 = Y
            mu0 = mu1
            mu1 = mu

            traj.append(Y)

        return Y, traj


# Copyright (c) 2023-2024 Salah Berra and contributors
# Distributed under the the GNU General Public License (See accompanying file LICENSE or copy
# at https://www.gnu.org/licenses/)

__author__ = "Salah Berra and contributors"
__copyright__ = "Copyright 2022-2023"
__license__ = "GNU GPL"
__version__ = "0.1.0"
__maintainer__ = "Salah Berra / Mohamed Nennouche"
__email__ = "."
__status__ = "dev"

# Note that each Alghorithm import from method module and run in the main code Iterative.
from .methods import *
from .train_methods import *
from .utils import *


from .utils import device, decompose_matrix
import torch
import torch.nn as nn
from typing import Tuple, List

def train_model(
    model: torch.nn.Module, 
    optimizer: torch.optim.Optimizer, 
    loss_func: torch.nn.Module, 
    total_itr: int, 
    solution: torch.Tensor, 
    num_batch: int
) -> Tuple[torch.nn.Module, List[float]]:
    """
    Train the given model using the specified optimizer and loss function.

    Args:
        - model (torch.nn.Module): The model to be trained.
        - optimizer (torch.optim.Optimizer): The optimizer to use for training.
        - loss_func (torch.nn.Module): The loss function to use for training.
        - total_itr (int): The total number of iterations (generations) for training.
        - solution (torch.Tensor): The target solution tensor.
        - num_batch (int): The number of batches per iteration.

    Returns:
        Tuple[torch.nn.Module, List[float]]: The trained model and the list of loss values per iteration.

    """
    loss_gen = []
    for gen in range(total_itr):
        for i in range(num_batch):
            optimizer.zero_grad()
            x_hat, _ = model(gen + 1)
            loss = loss_func(x_hat, solution)
            loss.backward()
            optimizer.step()
            
            if i % 200 == 0:
                print("generation:", gen + 1, " batch:", i, "\t MSE loss:", loss.item())
        loss_gen.append(loss.item())
    return model, loss_gen


def evaluate_model(
    model: torch.nn.Module, 
    solution: torch.Tensor, 
    n: int, 
    bs: int, 
    total_itr: int, 
    device: torch.device = device
) -> List[float]:
    """
    Evaluate the model by calculating the mean squared error (MSE) between the solution and the model's predictions.

    Args:
        - model (torch.nn.Module): The model to be evaluated.
        - solution (torch.Tensor): The target solution tensor.
        - n (int): The dimension of the solution.
        - bs (int): The batch size.
        - total_itr (int): The total number of iterations for evaluation.
        - device (torch.device, optional): The device to run the evaluation on. Defaults to device.

    Returns:
        List[float]: A list of MSE values for each iteration.
    """
    norm_list = []
    for i in range(total_itr + 1):
        s_hat, _ = model(i)
        err = (torch.norm(solution.to(device) - s_hat.to(device)) ** 2).item() / (n * bs)
        norm_list.append(err)
    return norm_list


class SORNet(nn.Module):
    """Deep unfolded SOR with a constant step size.

    Args:
        - init_val_SORNet (float): Initial value for inv_omega.
        - A (torch.Tensor): Matrix A of the linear system.
        - H (torch.Tensor): Matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution of the linear equation.
        - device (torch.device): Device to run the model on ('cpu' or 'cuda').

    Attributes:
        - device (torch.device): Device to run the model on.
        - inv_omega (nn.Parameter): Inverse of the relaxation parameter omega.
        - A (torch.Tensor): Matrix A of the linear system.
        - D (torch.Tensor): Diagonal matrix D.
        - L (torch.Tensor): Lower triangular matrix L.
        - U (torch.Tensor): Upper triangular matrix U.
        - H (torch.Tensor): Matrix H.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution of the linear equation.
    
    Methods:
        forward(num_itr: int) -> Tuple[torch.Tensor, List[torch.Tensor]]:
            Perform forward pass of the SORNet model.
    """

    def __init__(self, init_val_SORNet: float, A: torch.Tensor, H: torch.Tensor, bs: int, y: torch.Tensor, device: torch.device = device):
        """
        Initialize the SORNet model.

        Args:
            - init_val_SORNet (float): Initial value for inv_omega.
            - A (torch.Tensor): Matrix A of the linear system.
            - H (torch.Tensor): Matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution of the linear equation.
            - device (torch.device): Device to run the model on ('cpu' or 'cuda').

        """
        super(SORNet, self).__init__()
        self.device = device
        self.inv_omega = nn.Parameter(torch.tensor(init_val_SORNet, device=device))

        A, D, L, U, _, _ = decompose_matrix(A)

        self.A = A.to(device)
        self.D = D.to(device)
        self.L = L.to(device)
        self.U = U.to(device)
        self.H = H.to(device)
        self.Dinv = torch.linalg.inv(D).to(device)
        self.bs = bs
        self.y = y.to(device)

    def forward(self, num_itr: int) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """
        Perform forward pass of the SORNet model.

        Args:
            num_itr (int): Number of iterations.

        Returns:
            - torch.Tensor: Output tensor.
            - list: List of intermediate results.

        """
        traj = []

        invM = torch.linalg.inv(self.inv_omega * self.D + self.L)
        s = torch.zeros(self.bs, self.H.size(1), device=self.device)
        traj.append(s)
        yMF = torch.matmul(self.y, self.H.T)
        s = torch.matmul(yMF, self.Dinv)

        for _ in range(num_itr):
            temp = torch.matmul(s, (self.inv_omega - 1) * self.D - self.U) + yMF
            s = torch.matmul(temp, invM)
            traj.append(s)

        return s, traj


class SOR_CHEBY_Net(nn.Module):
    """Deep unfolded SOR with Chebyshev acceleration.

    Args:
        - num_itr (int): Number of iterations.
        - init_val_SOR_CHEBY_Net_omega (float): Initial value for omega.
        - init_val_SOR_CHEBY_Net_gamma (float): Initial value for gamma.
        - init_val_SOR_CHEBY_Net_alpha (float): Initial value for inv_omega.
        - A (torch.Tensor): Matrix A of the linear system.
        - H (torch.Tensor): Matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution of the linear equation.
        - device (torch.device): Device to run the model on ('cpu' or 'cuda').

    Attributes:
        - device (torch.device): Device to run the model on.
        - gamma (nn.Parameter): Gamma parameter for each iteration.
        - omega (nn.Parameter): Omega parameter for each iteration.
        - inv_omega (nn.Parameter): Inverse of the relaxation parameter omega.
        - A (torch.Tensor): Matrix A of the linear system.
        - D (torch.Tensor): Diagonal matrix D.
        - L (torch.Tensor): Lower triangular matrix L.
        - U (torch.Tensor): Upper triangular matrix U.
        - H (torch.Tensor): Matrix H.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution of the linear equation.

    Methods:
        forward(num_itr: int) -> Tuple[torch.Tensor, list[torch.Tensor]]:
            Perform forward pass of the SOR_CHEBY_Net model.
    """

    def __init__(self, num_itr: int, init_val_SOR_CHEBY_Net_omega: float, init_val_SOR_CHEBY_Net_gamma: float, init_val_SOR_CHEBY_Net_alpha: float, A: torch.Tensor, H: torch.Tensor, bs: int, y: torch.Tensor, device: torch.device = device):
        """
        Initialize the SOR_CHEBY_Net model.

        Args:
            - num_itr (int): Number of iterations.
            - init_val_SOR_CHEBY_Net_omega (float): Initial value for omega.
            - init_val_SOR_CHEBY_Net_gamma (float): Initial value for gamma.
            - init_val_SOR_CHEBY_Net_alpha (float): Initial value for inv_omega.
            - A (torch.Tensor): Matrix A of the linear system.
            - H (torch.Tensor): Matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution of the linear equation.
            - device (torch.device): Device to run the model on ('cpu' or 'cuda').

        """
        super(SOR_CHEBY_Net, self).__init__()
        self.device = device
        self.gamma = nn.Parameter(init_val_SOR_CHEBY_Net_gamma * torch.ones(num_itr, device=device))
        self.omega = nn.Parameter(init_val_SOR_CHEBY_Net_omega * torch.ones(num_itr, device=device))
        self.inv_omega = nn.Parameter(torch.tensor(init_val_SOR_CHEBY_Net_alpha, device=device))

        A, D, L, U, _, _ = decompose_matrix(A)
        self.A = A
        self.D = D.to(device)
        self.L = L.to(device)
        self.U = U.to(device)
        self.H = H.to(device)
        self.Dinv = torch.linalg.inv(D).to(device)
        self.bs = bs
        self.y = y.to(device)

    def forward(self, num_itr: int) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """
        Perform forward pass of the SOR_CHEBY_Net model.

        Args:
            num_itr (int): Number of iterations.

        Returns:
            - torch.Tensor: Output tensor.
            - list: List of intermediate results.

        """
        traj = []

        invM = torch.linalg.inv(self.inv_omega * self.D + self.L)
        s = torch.zeros(self.bs, self.H.size(1), device=self.device)
        s_new = torch.zeros(self.bs, self.H.size(1), device=self.device)
        traj.append(s)
        yMF = torch.matmul(self.y, self.H.T)
        s = torch.matmul(yMF, self.Dinv)
        s_present = s
        s_old = torch.zeros_like(s_present)

        for i in range(num_itr):
            temp = torch.matmul(s, (self.inv_omega - 1) * self.D - self.U) + yMF
            s = torch.matmul(temp, invM)

            s_new = self.omega[i] * (self.gamma[i] * (s - s_present) + (s_present - s_old)) + s_old
            s_old = s
            s_present = s_new
            traj.append(s_new)

        return s_new, traj

# =====================================================================================

class AORNet(nn.Module):
    """Deep unfolded AOR with a constant step size.

    Args:
        - init_val_AORNet_r (float): Initial value for r.
        - init_val_AORNet_omega (float): Initial value for omega.
        - A (torch.Tensor): Matrix A of the linear system.
        - H (torch.Tensor): Matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution of the linear equation.
        - device (torch.device): Device to run the model on ('cpu' or 'cuda').

    Attributes:
        - device (torch.device): Device to run the model on.
        - r (nn.Parameter): Parameter r for AOR.
        - omega (nn.Parameter): Relaxation parameter omega.
        - A (torch.Tensor): Matrix A of the linear system.
        - D (torch.Tensor): Diagonal matrix D.
        - L (torch.Tensor): Lower triangular matrix L.
        - U (torch.Tensor): Upper triangular matrix U.
        - H (torch.Tensor): Matrix H.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution of the linear equation.
    
    Methods:
        forward(num_itr: int) -> Tuple[torch.Tensor, list[torch.Tensor]]:
            Perform forward pass of the AORNet model.
    """

    def __init__(self, init_val_AORNet_r: float, init_val_AORNet_omega: float, A: torch.Tensor, H: torch.Tensor, bs: int, y: torch.Tensor, device: torch.device = device):
        """
        Initialize the AORNet model.

        Args:
            - init_val_AORNet_r (float): Initial value for r.
            - init_val_AORNet_omega (float): Initial value for omega.
            - A (torch.Tensor): Matrix A of the linear system.
            - H (torch.Tensor): Matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution of the linear equation.
            - device (torch.device): Device to run the model on ('cpu' or 'cuda').

        """
        super(AORNet, self).__init__()
        self.device = device
        self.r = nn.Parameter(torch.tensor(init_val_AORNet_r, device=device))
        self.omega = nn.Parameter(torch.tensor(init_val_AORNet_omega, device=device))

        A, D, L, U, _, _ = decompose_matrix(A)
        self.A = A.to(device)
        self.D = D.to(device)
        self.L = L.to(device)
        self.U = U.to(device)
        self.H = H.to(device)
        self.Dinv = torch.linalg.inv(D).to(device)
        self.bs = bs
        self.y = y.to(device)

    def forward(self, num_itr: int) -> Tuple[torch.Tensor, list[torch.Tensor]]:
        """
        Perform forward pass of the AORNet model.

        Args:
            num_itr (int): Number of iterations.

        Returns:
            - torch.Tensor: Output tensor.
            - list: List of intermediate results.

        """
        traj = []

        invM = torch.linalg.inv(self.L - self.r * self.D)
        N = (1 - self.omega) * self.D + (self.omega - self.r) * self.L + self.omega * self.U
        s = torch.zeros(self.bs, self.H.size(1), device=self.device)
        traj.append(s)
        yMF = torch.matmul(self.y, self.H.T)
        s = torch.matmul(yMF, self.Dinv)

        for _ in range(num_itr):
            s = torch.matmul(s, torch.matmul(invM, N)) + torch.matmul(yMF, invM)
            traj.append(s)

        return s, traj


class RINet(nn.Module):
    """Deep unfolded Richardson iteration.

    Args:
        - init_val_RINet (float): Initial value for inv_omega.
        - A (torch.Tensor): Matrix A of the linear system.
        - H (torch.Tensor): Matrix H.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution of the linear equation.
        - device (torch.device): Device to run the model on ('cpu' or 'cuda').

    Attributes:
        - inv_omega (nn.Parameter): Inverse of the relaxation parameter omega.
        - A (torch.Tensor): Matrix A of the linear system.
        - D (torch.Tensor): Diagonal matrix D.
        - L (torch.Tensor): Lower triangular matrix L.
        - U (torch.Tensor): Upper triangular matrix U.
        - H (torch.Tensor): Matrix H.
        - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
        - bs (int): Batch size.
        - y (torch.Tensor): Solution of the linear equation.

    Methods:
        forward(num_itr: int) -> Tuple[torch.Tensor, list[torch.Tensor]]:
            Perform forward pass of the RINet model.
    """

    def __init__(self, init_val_RINet: float, A: torch.Tensor, H: torch.Tensor, bs: int, y: torch.Tensor, device: torch.device = device):
        """
        Initialize the RINet model.

        Args:
            - init_val_RINet (float): Initial value for inv_omega.
            - A (torch.Tensor): Matrix A of the linear system.
            - H (torch.Tensor): Matrix H.
            - bs (int): Batch size.
            - y (torch.Tensor): Solution of the linear equation.
            - device (torch.device): Device to run the model on ('cpu' or 'cuda').

        """
        super(RINet, self).__init__()
        self.inv_omega = nn.Parameter(torch.tensor(init_val_RINet, device=device))

        A, D, L, U, _, _ = decompose_matrix(A)
        self.A = A.to(device)
        self.D = D.to(device)
        self.L = L.to(device)
        self.U = U.to(device)
        self.H = H.to(device)
        self.Dinv = torch.linalg.inv(D).to(device)
        self.bs = bs
        self.y = y.to(device)

    def forward(self, num_itr: int) -> Tuple[torch.Tensor, list[torch.Tensor]]:
        """
        Perform forward pass of the RINet model.

        Args:
            num_itr (int): Number of iterations.

        Returns:
            - torch.Tensor: Output tensor.
            - list: List of intermediate results.

        """
        traj = []

        s = torch.zeros(self.bs, self.A.shape[0], device=self.device)
        traj.append(s)
        yMF = torch.matmul(self.y, self.H.T)
        s = torch.matmul(yMF, self.Dinv)

        for _ in range(num_itr):
            s = s + self.inv_omega * (yMF - torch.matmul(s, self.A))
            traj.append(s)

        return s, traj


from setuptools import setup, find_packages

setup(
    name='unfolding_linear',
    version='0.1.0',
    description='Deep unfolding of iterative methods to solve linear equations',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Salah Berra',
    author_email='salahberra39@gmail.com',
    url='https://github.com/Salahberra2022/deep_unfolding',
    license='GPLv3',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=[
        'numpy>=1.20.0',
        'torch>=2.2.0',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
    ],
)



# Copyright (c) 2023-2024 Salah Berra and contributors
# Distributed under the the GNU General Public License (See accompanying file LICENSE or copy
# at https://www.gnu.org/licenses/)

# This script contains the functions that will be used in the various modules of the dee_unfolding package.
import numpy as np
import math
import torch
from typing import Tuple

device = torch.device("cuda" if torch.cuda.is_available() else "cpu") # GPU, if not CPU
print(f"Code run on : {device}")

def generate_A_H_sol(n: int = 300, m: int = 600, seed: int = 12, bs: int = 10, device: torch.device = device) -> Tuple[np.ndarray, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Generate matrices A, H, and W, as well as the solution and y.

    Args:
        - n (int, optional): Number of rows. Defaults to 300.
        - m (int, optional): Number of columns. Defaults to 600.
        - seed (int, optional): Seed for random number generation. Defaults to 12.
        - bs (int, optional): Batch size. Defaults to 10.
        - device (torch.device, optional): Device to run the computations on. Defaults to the global 'device'.

    Returns:
        - Tuple[np.ndarray, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: 
            - Matrix A (square matrix) of shape (n, n)
            - Matrix H (random matrix) of shape (n, m)
            - Matrix W with diagonal eigenvalues of A of shape (n, n)
            - Solution tensor of shape (bs, n)
            - Tensor y resulting from solution@H of shape (bs, m)
    """
    np.random.seed(seed=seed)
    H = np.random.normal(0, 1.0 / math.sqrt(n), (n, m))
    A = np.dot(H, H.T)
    eig = np.linalg.eig(A)[0]  # Eigenvalues
    
    W = torch.Tensor(np.diag(eig)).to(device)  # Define the appropriate 'device'
    H = torch.from_numpy(H).float().to(device)  # Define the appropriate 'device'
    
    print("Condition number, min. and max. eigenvalues of A:")
    print(np.max(eig) / np.min(eig), np.max(eig), np.min(eig))
    
    solution = torch.normal(torch.zeros(bs, n), 1.0).to(device).detach()
    y = solution @ H.detach()
    
    return A, H, W, solution, y

def decompose_matrix(A: np.ndarray) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Decompose a given matrix into its diagonal, lower triangular, upper triangular components and their inverses.

    Args:
        - A (np.ndarray): Input square matrix to decompose.

    Returns:
        - Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: 
            - A (torch.Tensor): Original matrix converted to a torch tensor.
            - D (torch.Tensor): Diagonal matrix of A.
            - L (torch.Tensor): Lower triangular matrix of A.
            - U (torch.Tensor): Upper triangular matrix of A.
            - Dinv (torch.Tensor): Inverse of the diagonal matrix D.
            - Minv (torch.Tensor): Inverse of the matrix (D + L).
    """
    # Decomposed matrix calculations
    D = np.diag(np.diag(A))  # Diagonal matrix
    L = np.tril(A, -1)       # Lower triangular matrix
    U = np.triu(A, 1)        # Upper triangular matrix
    Dinv = np.linalg.inv(D)  # Inverse of the diagonal matrix
    invM = np.linalg.inv(D + L)  # Inverse of the matrix (D + L)

    # Convert to Torch tensors and move to device
    A = torch.Tensor(A).to(device)
    D = torch.Tensor(D).to(device)
    L = torch.Tensor(L).to(device)
    U = torch.Tensor(U).to(device)
    Dinv = torch.Tensor(Dinv).to(device)
    Minv = torch.Tensor(invM).to(device)
    
    return A, D, L, U, Dinv, Minv